/*
 *  Copyright (C) 2000-2001  The Exult Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#  include <csignal>
#  include <cstdio>
#  include <cstdlib>
#  include <cstring>
#  include <iostream>
#  include <vector>

extern "C" {
#include "Duke3d.h"
#include "pragmas.h"
};

#include "pentagram/databuf.h"

#define	TRAILING_VOC_SLOP 32
#define	LEADING_VOC_SLOP 32

#ifndef WAVE_FORMAT_PCM
#define WAVE_FORMAT_PCM 1
#endif

static size_t	wav_buffer_size = 0;
static uint8	*wav_buffer = 0;

using std::vector;

struct	Chunk
{
	sint32	length;
	uint8	*data;
	Chunk(sint32 l, uint8 *d) : length(l),data(d) {}
};

static	sint16 *resample_new(uint8 *sourcedata,
						sint32 sourcelen, sint32 &destlen,
						int current_rate, int wanted_rate);
static	sint16 *resample_new16(uint8 *sourcedata,
						sint32 sourcelen, sint32 &destlen,
						int current_rate, int wanted_rate);

static void decode_ADPCM_4(uint8* inBuf,	
						  int bufSize,				// Size of inbuf
						  uint8* outBuf,	// Size is 2x bufsize
						  int& reference,			// ADPCM reference value
						  int& scale);

static	uint8 *chunks_to_wave(vector<Chunk> &chunks, unsigned long &size);

//----- Utilities ----------------------------------------------------

/*
 * Class that performs cubic interpolation on integer data.
 * It is expected that the data is equidistant, i.e. all have the same
 * horizontal distance. This is obviously the case for sampled audio.
 */
class CubicInterpolator {
protected:
	int x0, x1, x2, x3;
	int a, b, c, d;
	
public:
	CubicInterpolator(int a0, int a1, int a2, int a3) : x0(a0), x1(a1), x2(a2), x3(a3)
	{
		updateCoefficients();
	}
	
	CubicInterpolator(int a1, int a2, int a3) : x0(2*a1-a2), x1(a1), x2(a2), x3(a3)
	{
		// We use a simple linear interpolation for x0
		updateCoefficients();
	}
	
	inline void feedData()
	{
		x0 = x1;
		x1 = x2;
		x2 = x3;
		x3 = 2*x2-x1;	// Simple linear interpolation
		updateCoefficients();
	}

	inline void feedData(int xNew)
	{
		x0 = x1;
		x1 = x2;
		x2 = x3;
		x3 = xNew;
		updateCoefficients();
	}
	
	/* t must be a 16.16 fixed point number between 0 and 1 */
	inline int interpolate(uint32 fp_pos)
	{
		int result = 0;
		int t = fp_pos >> 8;
		result = (a*t + b) >> 8;
		result = (result * t + c) >> 8;
		result = (result * t + d) >> 8;
		result = (result/3 + 1) >> 1;
		
		return result;
	}
		
protected:
	inline void updateCoefficients()
	{
		a = ((-x0*2)+(x1*5)-(x2*4)+x3);
		b = ((x0+x2-(2*x1))*6) << 8;
		c = ((-4*x0)+x1+(x2*4)-x3) << 8;
		d = (x1*6) << 8;
	}
};

#define VOC_FORMAT_8BIT			0
#define VOC_FORMAT_4BIT_COMP	1
#define VOC_FORMAT_3BIT_COMP	2
#define VOC_FORMAT_2BIT_COMP	3
#define VOC_FORMAT_16BIT		4

extern "C" uint8 *convert_VOC(uint8 *old_data,uint32 *visible_len, int desired_sample_rate)
{
	std::vector<Chunk> chunks;
	bool	last_chunk=false;
	sint32	len = 0;
	sint32	chunk_type = 0;
	sint32	next_chunk = 0x1A;

	long	sample_rate = 0;
	long	format = 0;
	long	channels = 1;
	long	bps = 8;

	DataSource		ds(old_data);
	

	while(!last_chunk)
	{
		ds.seek(next_chunk);
		chunk_type = ds.read1();

		if (chunk_type == 0) break;

		len = ds.read3();
		next_chunk = len + ds.getPos();

		
		switch(chunk_type)
		{
			case 1:	// Sound Data
				sample_rate = ds.read1();
				format = ds.read1();

				if (format != VOC_FORMAT_8BIT)
				{
					last_chunk = true;
					continue;
				}

				sample_rate=1000000/(256-sample_rate);
				channels = 1;
				bps = 8;

				//workaround here to exit this loop, it fixes start speech which was
				//causing this function to go exit too early.
				//last_chunk=true;

				break;

			case 2:	// Continue Sound Data
				break;

			case 9:	// Extended Sound Data
				sample_rate = ds.read4();	// 4
				bps = ds.read1();			// 5
				channels = ds.read1();		// 6
				format = ds.read1();		// 7
				ds.skip(5);					// 12

				if (channels != 1 || (bps != 8 && bps != 16) || 
					(format != VOC_FORMAT_8BIT && format != VOC_FORMAT_16BIT) ||
						(bps == 16 && format != VOC_FORMAT_16BIT) ||
						(bps == 8 && format != VOC_FORMAT_8BIT) )
				{
					last_chunk = true;
					continue;
				}

				//workaround here to exit this loop, it fixes start speech which was
				//causing this function to go exit too early.
				//last_chunk=true;

				break;
			default:
				continue;
		}

		uint8 *dec_data = ds.getPtr();
		sint32 dec_len = next_chunk - ds.getPos();

		if (dec_len <= 0)
		{
			break;
		}

		// 

		// Our input is 8 bit mono unsigned; but want to output 16 bit stereo signed.
		// In addition, the rates don't match, we have to upsample.
		// New code: Do it all in one step with cubic interpolation

		sint16 *new_data = 0;
		sint32 new_size = 0;
		if (format == VOC_FORMAT_8BIT)
			new_data = resample_new(dec_data, dec_len, new_size, sample_rate, desired_sample_rate);
		else if (format == VOC_FORMAT_16BIT)
			new_data = resample_new16(dec_data, dec_len, new_size, sample_rate, desired_sample_rate);


		if (new_size) chunks.push_back(Chunk(new_size,(uint8 *)new_data));
		else if (new_data) delete [] new_data;

	}
	

	return chunks_to_wave(chunks, *visible_len);
}

static	sint16 *resample_new(uint8 *src,
						sint32 sourcelen, sint32 &size,
						int rate, int wanted_rate)
{
	int fp_pos = 0;
	//int fp_speed = (1 << 16) * rate / wanted_rate;
	int fp_speed = scale((1 << 16),rate,wanted_rate);
	size = sourcelen;

	// adjust the magnitudes of size and rate to prevent division error
//	while (size & 0xFFFF0000)
//		size >>= 1, rate = (rate >> 1) + 1;
	
	// Compute the output size (times 4 since it is 16 stereo)
	size = scale(size,wanted_rate,rate) << 1;

	sint16 *new_data = new sint16 [size];
	sint16 *data = new_data;
	uint8 *src_end = src + sourcelen;

	int result;
	
	// Compute the initial data feed for the interpolator. We don't simply
	// shift by 8, but rather duplicate the byte, this way we cover the full
	// range. Probably doesn't make a big difference, listening wise :-)
	int a = *(src+0); a |= (a << 8);
	int b = *(src+1); b |= (b << 8);
	int c = *(src+2); c |= (c << 8);
	
	// We divide the data by 2, to prevent overshots. Imagine this sample pattern:
	// 0, 65535, 65535, 0. Now you want to compute a value between the two 65535.
	// Obviously, it will be *bigger* than 65535 (it can get to about 80,000).
	// It is possibly to clamp it, but that leads to a distored wave form. Compare
	// this to turning up the volume of your stereo to much, it will start to sound
	// bad at a certain level (depending on the power of your stereo, your speakers 
	// etc, this can be quite loud, though ;-). Hence we reduce the original range.
	// A factor of roughly 1/1.2 = 0.8333 is sufficient. Since we want to avoid 
	// floating point, we approximate that by 27/32
	#define RANGE_REDUX(x)	(((x) * 27) >> 5)
//	#define RANGE_REDUX(x)	((x) >> 1)
//	#define RANGE_REDUX(x)	((x) / 1.2)

#ifdef USE_CUBIC
	CubicInterpolator	interp(RANGE_REDUX(a), RANGE_REDUX(b), RANGE_REDUX(c));
#endif
	
	do {
		do {
			// Convert to signed data
#ifdef USE_CUBIC
			result = interp.interpolate(fp_pos) - 32768;
#else
			result = a - 32768;
#endif

			// Enforce range in case of an "overshot". Shouldn't happen since we
			// scale down already, but safe is safe.
			if (result < -32768)
				result = -32768;
			else if (result > 32767)
				result = 32767;
	
			*data++ = result;
	
			fp_pos += fp_speed;
		} while (!(fp_pos & 0xFFFF0000));
		src++;
		fp_pos &= 0x0000FFFF;
		

		a = b;
		b = c;

		if (src+2 < src_end) {
			c = *(src+2);
			c |= (c << 8);
#ifdef USE_CUBIC
			interp.feedData(RANGE_REDUX(c));
#endif
		} else {
#ifdef USE_CUBIC
			interp.feedData();
#endif
		}

	} while (src < src_end);
	
	return new_data;
}

static	sint16 *resample_new16(uint8 *src8,
						sint32 sourcelen, sint32 &size,
						int rate, int wanted_rate)
{
	int fp_pos = 0;
	int fp_speed = scale((1 << 16),rate,wanted_rate);
	sint16 *src = (sint16*) src8;
	size = sourcelen;

	// adjust the magnitudes of size and rate to prevent division error
	//while (size & 0xFFFF0000)
	//	size >= 2, rate = (rate >> 1) + 1;
	
	// Compute the output size (times 2 since it is 16 stereo)
	size = scale(size,wanted_rate,rate);

	sint16 *new_data = new sint16 [size];
	sint16 *data = new_data;
	sint16 *src_end = (sint16*)(src8 + sourcelen);

	int result;
	
	int a = ((int) *(src+0)) + 32768;
	int b = ((int)*(src+1)) + 32768;
	int c = ((int)*(src+2)) + 32768;
	
	// We divide the data by 2, to prevent overshots. Imagine this sample pattern:
	// 0, 65535, 65535, 0. Now you want to compute a value between the two 65535.
	// Obviously, it will be *bigger* than 65535 (it can get to about 80,000).
	// It is possibly to clamp it, but that leads to a distored wave form. Compare
	// this to turning up the volume of your stereo to much, it will start to sound
	// bad at a certain level (depending on the power of your stereo, your speakers 
	// etc, this can be quite loud, though ;-). Hence we reduce the original range.
	// A factor of roughly 1/1.2 = 0.8333 is sufficient. Since we want to avoid 
	// floating point, we approximate that by 27/32
	#define RANGE_REDUX(x)	(((x) * 27) >> 5)
//	#define RANGE_REDUX(x)	((x) >> 1)
//	#define RANGE_REDUX(x)	((x) / 1.2)

#ifdef USE_CUBIC
	CubicInterpolator	interp(RANGE_REDUX(a), RANGE_REDUX(b), RANGE_REDUX(c));
#endif
	
	do {
		do {
			// Convert to signed data
#ifdef USE_CUBIC
			result = interp.interpolate(fp_pos) - 32768;
#else
			result = a - 32768;
#endif


			// Enforce range in case of an "overshot". Shouldn't happen since we
			// scale down already, but safe is safe.
			if (result < -32768)
				result = -32768;
			else if (result > 32767)
				result = 32767;
	
			*data++ = result;
	
			fp_pos += fp_speed;
		} while (!(fp_pos & 0xFFFF0000));
		src++;
		fp_pos &= 0x0000FFFF;
		
		a = b;
		b = c;

		if (src+2 < src_end) {
			c = *(src+2) + 32768;
#ifdef USE_CUBIC
			interp.feedData(RANGE_REDUX(c));
#endif
		} else {
#ifdef USE_CUBIC
			interp.feedData();
#endif
		}

	} while (src < src_end);
	
	return new_data;
}

static	uint8 *chunks_to_wave(vector<Chunk> &chunks, unsigned long &size)
{
	sint32	aggregate_length=0;
	sint32	working_offset=0;
	
	for(std::vector<Chunk>::iterator it=chunks.begin();
		it!=chunks.end(); ++it)
		{
		aggregate_length+=it->length;
		}
	if (!aggregate_length) 
	{
		size = 0;
		return 0;
	}

	if (wav_buffer_size < (aggregate_length+0x40))
	{
		delete [] wav_buffer;
		wav_buffer = new uint8[aggregate_length+0x40];
		wav_buffer_size = aggregate_length+0x40;
	}

	DataSource writer((char *) wav_buffer, wav_buffer_size);

	// Write wav header
	writer.write  ("RIFF", 4);				// RIFF
	writer.write4 (aggregate_length+0x24);	// Filesize - 4
	writer.write  ("WAVE", 4);				// Type

	writer.write  ("fmt ", 4);				// fmt Chunk
	writer.write4 (0x10);					// Chunk size
	writer.write2 (WAVE_FORMAT_PCM);		// wFormatTag 
	writer.write2 (1);						// nChannels (2 channels, stereo)
	writer.write4 (MixRate);				// nSamplesPerSec
	writer.write4 (MixRate*2);				// nAvgBytesPerSec
	writer.write2 (2);						// nBlockAlign (nChannels*wBitsPerSample/8)
	writer.write2 (16);						// wBitsPerSample

	writer.write  ("data", 4);				// data chunk
	writer.write4 (aggregate_length);		// buffer size

	{
		for(std::vector<Chunk>::iterator it=chunks.begin();
			it!=chunks.end(); ++it)
			{
			//memcpy(unified_block+working_offset,it->data,it->length);
			writer.write((char*) it->data,it->length);
			working_offset+=it->length;
			delete [] it->data; it->data=0; it->length=0;
			}
	}

	size = writer.getPos();
	return wav_buffer;
}
